/*
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *    Timon Veenstra <monezz@gmail.com> - initial API and implementation and/or initial documentation
 */
package eu.agrosense.client.grid;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LinearRing;
import com.vividsolutions.jts.geom.Polygon;
import com.vividsolutions.jts.geom.impl.CoordinateArraySequence;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.Point;
import java.awt.image.DataBuffer;
import java.awt.image.WritableRaster;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import nl.cloudfarming.client.geoviewer.Geographical;
import nl.cloudfarming.client.model.AbstractEntity;
import nl.cloudfarming.client.model.AgroURI;
import nl.cloudfarming.client.model.EnterpriseIdProvider;
import nl.cloudfarming.client.model.ItemIdType;
import org.geotools.coverage.CoverageFactoryFinder;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridCoverageFactory;
import org.geotools.geometry.jts.JTSFactoryFinder;
import org.opengis.geometry.BoundingBox;
import org.openide.util.Exceptions;

/**
 * A grid is a north/south oriented raster of cells with a fixed with/height.
 * The bounds of the grid are geo referenced, ie they have coordinates. The Grid
 * class is a wrapper for GridCoverage2D with some additional information about
 * the grid.
 *
 * @author Timon Veenstra <monezz@gmail.com>
 */
public class Grid extends AbstractEntity implements Geographical<GridCoverage2D> {

    public static final String PROP_GRID = "grid";
    private static final String PROP_SOURCES = "sources";
    
    
    // variables required initialisation before functional usage of this class
    private BoundingBox boundingBox;
    private Geometry boundingBoxGeometry;
    private WritableRaster raster;
    
    private Future<BoundingBox> boundingBoxFuture;
    
    
    private Set<AgroURI> sources;
    private boolean dirty = true;
    private GridCoverage2D grid;
    public static final ItemIdType ITEM_ID_TYPE = ItemIdType.GRD;
    public static final String EXT = "grd";
    public static final String MIME_TYPE = "application/x-agrosense-grid";
    private final GridCellSize gridCellSize;
    private static final Logger LOGGER = Logger.getLogger(Grid.class.getName());

    private void checkState(){
        if (boundingBox == null || boundingBoxGeometry == null || raster == null && boundingBoxFuture != null){
            try {
                setBoundingBox(boundingBoxFuture.get());
            } catch (InterruptedException | ExecutionException ex) {
                Exceptions.printStackTrace(ex);
            }
        }
        if (boundingBox == null || boundingBoxGeometry == null || raster == null){
            throw new IllegalStateException("Either one of the variables boundingBox, boundingBoxGeometry or raster has not been initialized");
        }
    }


    public Grid(EnterpriseIdProvider eip, GridCellSize gridCellSize, Future<BoundingBox> boundingBoxFuture) {
        super(eip, ItemIdType.GRD);
        this.boundingBoxFuture = boundingBoxFuture;
        this.gridCellSize = gridCellSize;
    }
    
    
    
    public Grid(EnterpriseIdProvider eip, GridCellSize gridCellSize, BoundingBox boundingBox) {
        super(eip, ITEM_ID_TYPE);
        this.gridCellSize = gridCellSize;
        setBoundingBox(boundingBox);
    }

    public void setRasterValue(Coordinate coordinate, int value) {
        checkState();
        if (!boundingBox.contains(coordinate.x, coordinate.y)) {
            throw new IllegalArgumentException("Can only add raster values for coordinates (" + coordinate + ") within the bounding box (" + boundingBox + ")");
        }

        Point p = this.gridCellSize.getCell(boundingBox, coordinate);
        LOGGER.log(Level.FINEST, "raster value set to {0} on cell {1}", new Object[]{value, p});        
        
        this.raster.setSample(p.x, p.y, 0, value);
        dirty = true;
    }

    /**
     * copy constructor.
     *
     * Creates a new grid based on the provided grid.
     *
     * //TODO use empty internal grid
     *
     * @param other
     */
    public Grid(Grid other) {
        this(other.getURI(), other.getGridCellSize(), other.boundingBox);
    }

    public GridCellSize getGridCellSize() {
        return gridCellSize;
    }
    
    public BoundingBox getInternalBoundingBox(){
        return boundingBox;
    }

    @Override
    public Geometry getBoundingBox() {
        checkState();
        return boundingBoxGeometry;
    }

    /**
     * getting the internal value raster for a grid will mark it as dirty.
     * 
     * @return 
     */
    public WritableRaster getRaster() {
        checkState();
        this.dirty = true;
        return raster;
    }
    
    

    /**
     * Get the grid. When the internal raster data is changed, the old grid is
     * returned. When on the EDT, the grid will be updated on a background
     * thread. Otherwise the method will wait for the result. Listen to property
     * change listener to get notified.
     *
     * @return
     */
    public GridCoverage2D getGrid() {
        checkState();
        if (dirty && SwingUtilities.isEventDispatchThread()) {
            if (grid == null){
                grid = getDummyRaster();
            }
            new SwingWorker() {
                @Override
                protected Object doInBackground() throws Exception {
                    generateGrid();
                    return null;
                }
            }.execute();
            //NOT ON THE EDT
        } else if (dirty) {
            generateGrid();
        }
        return grid;
    }
    
    protected  GridCoverage2D getDummyRaster() {
        checkState();
        WritableRaster dummyRaster = WritableRaster.createBandedRaster(DataBuffer.TYPE_INT, 1,1,1, new Point(0,0));
        dummyRaster.setSample(0,0,0,Integer.MIN_VALUE);
        GridCoverageFactory factory = CoverageFactoryFinder.getGridCoverageFactory(null);
        return factory.create("name", dummyRaster, boundingBox);        
    }    

    public final void setBoundingBox(BoundingBox boundingBox) {
        this.boundingBox = boundingBox;
        LOGGER.finest("Initializing bounding box for Grid");
        Coordinate[] coordinates = new Coordinate[5];

        coordinates[0] = new Coordinate(boundingBox.getMinX(), boundingBox.getMinY());
        coordinates[1] = new Coordinate(boundingBox.getMaxX(), boundingBox.getMinY());
        coordinates[2] = new Coordinate(boundingBox.getMaxX(), boundingBox.getMaxY());
        coordinates[3] = new Coordinate(boundingBox.getMinX(), boundingBox.getMaxY());
        coordinates[4] = new Coordinate(boundingBox.getMinX(), boundingBox.getMinY());

        GeometryFactory geometryFactory = JTSFactoryFinder.getGeometryFactory(null);
        LinearRing linearRing = new LinearRing(new CoordinateArraySequence(coordinates), geometryFactory);
        boundingBoxGeometry = new Polygon(linearRing, null, geometryFactory);
        
        Dimension d = gridCellSize.getDimension(boundingBox);
        
        int bands = 1;
        Point location = new Point(0, 0);
        this.raster = WritableRaster.createBandedRaster(DataBuffer.TYPE_INT, d.width, d.height, bands, location);
        LOGGER.log(Level.FINEST, "created raster with dimension{0}", d);
        // fill raster with min values
        for (int iw = 0; iw < d.width; iw++) {
            for (int ih = 0; ih < d.height; ih++) {
                raster.setSample(iw, ih, 0, Integer.MIN_VALUE);
            }
        }        
    }

    
    
    /**
     * NEVER EXECUTE ON THE EDT!
     */
    private void generateGrid() {
        checkState();
        assert !SwingUtilities.isEventDispatchThread() : "generateGrid should never be invoked on the EDT!";
        GridCoverageFactory factory = CoverageFactoryFinder.getGridCoverageFactory(null);
        setGrid(factory.create(Grid.class.getName(), raster, boundingBox));
        dirty = false;
    }

    private void setGrid(GridCoverage2D grid) {
        GridCoverage2D old = this.grid;
        this.grid = grid;
        getPropertyChangeSupport().firePropertyChange(PROP_GRID, old, grid);
    }

    @Override
    public com.vividsolutions.jts.geom.Point getCentroid() {
        checkState();
        return this.boundingBoxGeometry.getCentroid();
    }

    @Override
    public GridCoverage2D getRenderObject(Envelope envelope) {
        checkState();
        return getGrid();
    }

    @Override
    public String getTooltipText() {
        return null;
    }

    @Override
    public String getIconLabel() {
        return null;
    }

    @Override
    public Image getIcon() {
        return null;
    }
}
