/*
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *    Timon Veenstra <monezz@gmail.com> - initial API and implementation and/or initial documentation
 */
package eu.agrosense.client.grid.converter;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import eu.agrosense.client.grid.Grid;
import eu.agrosense.client.grid.GridCellSize;
import eu.agrosense.client.grid.GridConverter;
import eu.agrosense.client.grid.GridProvider;
import eu.agrosense.client.grid.TemporaryGrid;
import java.awt.Point;
import java.awt.image.WritableRaster;
import nl.cloudfarming.client.model.EnterpriseIdProvider;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.lookup.ServiceProvider;

/**
 *
 * @author Timon Veenstra <monezz@gmail.com>
 */
@ServiceProvider(service=GridConverter.class)
@NbBundle.Messages(
        {"and grid converter display name=Add Grid Converter"
        ,"and grid converter description=The Add Grid converter, takes two grids and adds the values of all overlapping grid cells into a new grid."})
public class AddConverter implements GridConverter{

    @Override
    public String getDisplayName() {
        return Bundle.and_grid_converter_display_name();
    }

    @Override
    public String getDescription() {
        return Bundle.and_grid_converter_description();
    }

    @Override
    public int getSourceGridMax() {
        return 2;
    }

    @Override
    public Grid convert(Grid... grids) {
        assert grids.length >= 2:"Converter requires at least 2 grid to operate";
        assert !(grids[0] instanceof TemporaryGrid && grids[1] instanceof TemporaryGrid):"Provided grids cannot both be a TemporaryGrid";
        
        EnterpriseIdProvider eip = (grids[0] instanceof TemporaryGrid)?grids[1].getURI():grids[0].getURI();
        Geometry overlap = grids[0].getBoundingBox().intersection(grids[1].getBoundingBox());
        
        Coordinate topLeft = new Coordinate(overlap.getEnvelopeInternal().getMinX(), overlap.getEnvelopeInternal().getMinY());
        Coordinate bottomRight = new Coordinate(overlap.getEnvelopeInternal().getMaxX(), overlap.getEnvelopeInternal().getMaxY());
        
        GridProvider gridProvider = Lookup.getDefault().lookup(GridProvider.class);
        // take smallest gridCellSize of the 2
        GridCellSize cellSize = (grids[0].getGridCellSize().compareTo(grids[1].getGridCellSize())<=0)?grids[0].getGridCellSize():grids[1].getGridCellSize();
        
        Grid resultGrid = gridProvider.getDefaultGridSection(eip, topLeft, bottomRight, cellSize);
        
        //process values of grid 0
        // determine box of samples we can use for evaluation
        Point topLeftGrid0 = cellSize.getCell(grids[0].getInternalBoundingBox(), topLeft);
        Point bottomRightGrid0 = cellSize.getCell(grids[0].getInternalBoundingBox(), bottomRight);
        int widthGrid0 = topLeftGrid0.x - bottomRightGrid0.x;
        int heightGrid0 = topLeftGrid0.y - bottomRightGrid0.y;
        
        int[] grid0Samples = grids[0].getRaster().getSamples(topLeftGrid0.x,topLeftGrid0.y, widthGrid0, heightGrid0, 0, (int[])null);
        
        // process values of grid 1
        // determine box of samples we can use for evaluation
        Point topLeftGrid1 = cellSize.getCell(grids[1].getInternalBoundingBox(), topLeft);
        Point bottomRightGrid1= cellSize.getCell(grids[1].getInternalBoundingBox(), bottomRight);
        int widthGrid1 = topLeftGrid1.x - bottomRightGrid1.x;
        int heightGrid1 = topLeftGrid1.y - bottomRightGrid1.y;
        
        int[] grid1Samples = grids[1].getRaster().getSamples(topLeftGrid1.x,topLeftGrid1.y, widthGrid1, heightGrid1, 0, (int[])null); 
        
        assert grid0Samples.length==grid1Samples.length:"Grid 0 and 1 samples from intersection should have same sample amount";
        
        int[] gridResultSamples = new int[grid0Samples.length];
        for (int i=0;i<gridResultSamples.length;i++){
            gridResultSamples[i] = (grid0Samples[i]==Integer.MIN_VALUE||grid1Samples[i]==Integer.MIN_VALUE)?Integer.MIN_VALUE:grid0Samples[i]+grid1Samples[i];
        }
        
        WritableRaster resultRaster = resultGrid.getRaster();
        resultRaster.setSamples(resultRaster.getMinX(), resultRaster.getMinY(), resultRaster.getWidth(), resultRaster.getHeight(), 0, gridResultSamples);
        
        
        // pixel location = x* y+width
        
        return resultGrid;
    }
    
}
